﻿Option Explicit On
Option Strict On

''' <summary>
''' CommandParameter() is the base class for each of the variable classes (AnalogInput,
''' AnalogOutput, DigitalInput, DigitalOutput).
''' </summary>
''' <remarks>Copyright (c) 2015   Vanderhaegen Bart</remarks>
Public Class CommandParameter

    ' Class variables
    Private parameter As String
    Public Property isString As Boolean
    Public Property minChannel As Integer
    Public Property maxChannel As Integer
    Private Property commandDataType As dataType

    ' Enumeration dataType
    Public Enum dataType
        [STRING]
        [INTEGER]
    End Enum



    ''' <summary>
    ''' The class constructor. Should be overriden
    ''' </summary>
    ''' <param name="parameter">The parameter to assign to this variable</param>
    ''' <remarks></remarks>
    Public Sub New(parameter As String)
        Me.parameter = parameter
    End Sub




    ''' <summary>
    ''' isValid() returns true if the contents of the variable corresponds with
    ''' the expected datatype.
    ''' </summary>
    ''' <returns>True if the variable corresponds with the datatype</returns>
    ''' <remarks></remarks>
    Public Function isValid() As Boolean
        Select Case commandDataType
            Case dataType.STRING
                ' String expected. Check if not empty.
                Return parameter <> String.Empty

            Case dataType.INTEGER
                ' Integer expected. Check bounds.
                Return fitBounds()

            Case Else
                ' This case should never happen
                ' !!! define other dattypes
                Return False
        End Select
    End Function




    ''' <summary>
    ''' fitBounds() checks if the integer value lies within the expected range.
    ''' The function returns false if the current datatype isn't an integer.
    ''' </summary>
    ''' <returns>True if the value lies within the specified range.</returns>
    ''' <remarks></remarks>
    Private Function fitBounds() As Boolean
        Dim iValue As Integer
        If Integer.TryParse(parameter, iValue) Then
            Return iValue >= minChannel And iValue <= maxChannel
        Else
            Return False
        End If
    End Function




    ''' <summary>
    ''' setBounds() will set the bounds for the class. This only needs to be
    ''' done if the class is going to be used for Integer values.
    ''' </summary>
    ''' <param name="minChannel">minimal allowed value</param>
    ''' <param name="maxChannel">maximal allowed value</param>
    ''' <remarks></remarks>
    Public Sub setBounds(minChannel As Integer, maxChannel As Integer)
        Me.minChannel = minChannel
        Me.maxChannel = maxChannel
    End Sub




    ''' <summary>
    ''' setType() is a setter to select the desired datatype.
    ''' </summary>
    ''' <param name="type">One of the enumerated possibilities:
    ''' dataType.STRING or dataTYPE.integer</param>
    ''' <remarks></remarks>
    Public Sub setType([type] As dataType)
        commandDataType = type
    End Sub





    ''' <summary>
    ''' toInt() converts the value of the object to an integer, if applicable.
    ''' The function returns 0 in case of an error. This should never happen.
    ''' </summary>
    ''' <returns>The integer value</returns>
    ''' <remarks></remarks>
    Public Function toInt() As Integer
        Try
            Return Integer.Parse(parameter)
        Catch ex As Exception
            ' The exception should never occur
            Return 0
        End Try
    End Function




    ''' <summary>
    ''' toString() returns the string value of object.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function toString() As String
        Return parameter
    End Function
End Class
