﻿Option Explicit On
Option Strict On

Imports System.Text
Imports System.Reflection


''' <summary>
''' ServerFunctions() defines the commands that can be understood by the server. The class contains 
''' a separate function for each of the commands. During runtime the class will analyse itself (by
''' using .NET reflection) and and create a list (functionList) with an overview of all allowed 
''' commands.
''' </summary>
''' <remarks>Copyright (c) 2015   Vanderhaegen Bart</remarks>
Public Class ServerFunctions

    ' Class definitions
    Private mK8055 As K8055                                 ' K8055 object
    Public Property [readOnly] As Boolean                   ' is K8055 in readonly mode?
    Private functionList As New List(Of ServerFunction)     ' list with all functions

    ' SERVERFUNCTION_PREFIX defines how all functions with server commands should be named.
    ' By default each function should start by the Cmd_. 
    ' Example:
    ' If you would like to add a function called STOP, you would need to add a function
    ' named Cmd_Stop() to this class.
    Private Const SERVERFUNCTION_PREFIX = "Cmd_"



    ''' <summary>
    ''' Class constructor. Always pass an instance of the K8055 object.
    ''' </summary>
    ''' <param name="interface">The k8055 interface object</param>
    ''' <remarks></remarks>
    Public Sub New(ByRef [interface] As K8055)
        mK8055 = [interface]

        ' Start the .NET reflection process
        defineServerFunctions()
    End Sub




    ''' <summary>
    ''' canReadWrite() returns true if the K8055 is in readwrite mode.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Protected Friend Function canReadWrite() As Boolean
        Return (Not [readOnly])
    End Function





#Region "Reflection"
    ''' <summary>
    ''' defineServerFunctions() will diagnose the class and lists the server functions. Next it will
    ''' check if each function has the required response type, and the required parameter types.
    ''' The function can throw errors but this will only happen if the programmer tried adding a 
    ''' function that wasn't complient.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub defineServerFunctions()
        For Each curFunction As MethodInfo In Me.GetType.GetMethods
            ' Loop through each method of the class
            If isCmdFunction(curFunction) Then
                ' The function has the right name
                If hasRequiredReturnType(curFunction) Then
                    ' The function has the expected return type
                    If hasValidParameters(curFunction) Then
                        ' The function had the expected parameters
                        addFunction(stripFunctionName(curFunction.Name), getParameterTypes(curFunction))
                    Else
                        ' The function has one or more invalid parameters (should be objects inherenting
                        ' from CommandParameter !!!)
                        Throw New Exception("Inner error: the class K8055_Interpreter contains an user-defined " & _
                                            "function with a parameter type that does not inherit from 'CommandParameter'")
                    End If

                Else
                    ' The function has an invalid output type (should be a ServerResponse object)
                    Throw New Exception("Inner error: the class K8055_Interpreter contains an user-defined" & _
                                        "function with an invalid return type.")
                End If 'hasRequiredReturnType(curFunction) Then
            End If 'isFunctionDefined(curFunction) Then
        Next ' For Each curFunction As MethodInfo In Me.GetType.GetMethods
    End Sub




    ''' <summary>
    ''' getParameterTypes() reads all parameter types for the specified function and returns
    ''' them in an array of System.Type(). 
    ''' </summary>
    ''' <param name="myFunction">The function to investigate</param>
    ''' <returns>An array with all types</returns>
    ''' <remarks></remarks>
    Private Function getParameterTypes(myFunction As MethodInfo) As Type()
        Dim myParams() As ParameterInfo = myFunction.GetParameters()
        Dim res(myParams.Count - 1) As System.Type

        If myParams.Count = 0 Then
            ' There are no parameters - return empty.
            Return Nothing
        Else
            ' there are one or more parameters. Loop through each one of them
            For i As Integer = 0 To myParams.Count - 1
                res(i) = myParams(i).ParameterType
            Next
            Return res
        End If
    End Function




    ''' <summary>
    ''' hasValidParameters() checks if the specified function only has parameters that
    ''' are classes which are derived from 'CommandParameter'.
    ''' </summary>
    ''' <param name="myFunction">The function to check</param>
    ''' <returns>True if all parameters are valid</returns>
    ''' <remarks></remarks>
    Private Function hasValidParameters(myFunction As MethodInfo) As Boolean
        Dim myParams() As ParameterInfo = myFunction.GetParameters()

        Dim res As Boolean = True
        For i As Integer = 0 To myParams.Count - 1
            ' For each parameter, compare BaseClass against CommandParameter
            res = res And (myParams(i).ParameterType.BaseType = GetType(CommandParameter))
        Next
        Return res
    End Function




    ''' <summary>
    ''' isCmdFunction() checks if the given function names starts with the
    ''' expected prefix, as defined by the constant SERVERFUNCTION_PREFIX.
    ''' </summary>
    ''' <param name="myFunction">Function to check</param>
    ''' <returns>True if the function starts with the expected prefix</returns>
    ''' <remarks></remarks>
    Private Function isCmdFunction(myFunction As MethodInfo) As Boolean
        Return myFunction.Name.ToUpper().Substring(0, SERVERFUNCTION_PREFIX.Length) =
            SERVERFUNCTION_PREFIX.ToUpper()
    End Function




    ''' <summary>
    ''' hasRequiredReturnType() checks if the given function has the
    ''' expected return type.
    ''' </summary>
    ''' <param name="myFunction">Function to check</param>
    ''' <returns>True if the function has the expected return type</returns>
    ''' <remarks></remarks>
    Private Function hasRequiredReturnType(myFunction As MethodInfo) As Boolean
        Return myFunction.ReturnType = GetType(ServerResponse)
    End Function




    ''' <summary>
    ''' stripFunctionName() extracts the prefix from the function name. 
    ''' 'Cmd_SetAllDigital' will become 'SetAllDigital'. Thus the function will
    ''' determine the real command that can be called from the client.
    ''' </summary>
    ''' <param name="functionName">the complete function name</param>
    ''' <returns>The command without prefix.</returns>
    ''' <remarks></remarks>
    Private Function stripFunctionName(functionName As String) As String
        Return functionName.Substring(SERVERFUNCTION_PREFIX.Length)
    End Function




    ''' <summary>
    ''' addFunction() adds the function to the list with known functions.
    ''' </summary>
    ''' <param name="functionName">The name of the function (without the
    ''' function prefix).</param>
    ''' <param name="functionType">An array with all required datatypes.
    ''' The variable might be set to null/nothing if there aren't any.</param>
    ''' <remarks></remarks>
    Private Sub addFunction(functionName As String, functionType As Type())
        functionList.Add(New ServerFunction(functionName, functionType))
    End Sub




    ''' <summary>
    ''' executeServerCommand() will execute a server command. This function will be executed
    ''' after the reception of each new command from a client. 
    ''' </summary>
    ''' <param name="instruction">The instruction code</param>
    ''' <param name="parameterValue">All parameter information (in a single string)</param>
    ''' <returns>The response string that will be replied straight to the client.</returns>
    ''' <remarks></remarks>
    Public Function executeServerCommand(ByVal instruction As String,
                                      ByVal parameterValue As String) As String

        ' First put the instruction in uppercase
        instruction = instruction.ToUpper()

        ' Loop through all defined functions
        For Each [function] As ServerFunction In functionList
            If [function].getName = instruction Then
                ' Function was found

                ' Define an empty CommandParameter array to be passed 
                Dim dParameter(0) As CommandParameter

                ' Check if the parameters are matching.
                ' !!! dParameter is passed as byRef.
                If [function].parameterMatches(parameterValue, dParameter) Then
                    ' All parameters matches with the expected parameter types
                    Dim response As ServerResponse

                    Try
                        If [function].requiresParameter Then
                            ' The function requires one or more parameters
                            response = DirectCast(CallByName(Me,
                                                    String.Format("Cmd_{0}", instruction),
                                                    CallType.Method,
                                                    dParameter), ServerResponse)

                        Else
                            ' The function doesn't require any parameters
                            response = DirectCast(CallByName(Me,
                                                    String.Format("Cmd_{0}", instruction),
                                                    CallType.Method), ServerResponse)
                        End If '[function].requiresParameter Then

                        ' Return the +OK servercode to the client
                        Return String.Format("+OK {0} {1}", response.returnCode, response.message)

                    Catch e As Exception
                        ' The function could not be executed. This case should never happen. This
                        ' would only happen in case of a programming error.
                        Return String.Format("-ERR A serious server-error occured")
                    End Try

                Else
                    ' There is a mismatch between actual and expected parameters
                    Return String.Format("-ERR Invalid parameter")
                End If '[function].parameterMatches(parameterValue, dParameter) Then
            End If
        Next

        Return ("-ERR Unknown function")
    End Function
#End Region




    ' Use the class ServerFunctions() to define all of your server functions (the functions that can 
    ' be called by the network clients). Each function you define here will automatically be 
    ' recognized. There are a few considerations:
    '
    ' 1) A function name should always start with Cmd_
    ' 2) The return type should be of the type ServerResponse
    ' 3) A function can have multiple parameters. Each parameter must inherit from 'CommandParameter'
    ' 4) Each function should be executed as soon as possible. It may NOT contain any delaying functions,
    '    like thread.sleep().
    ' 5) The response codes are defines in the class Globals.vb.
    ' 
#Region "Functions"
    ' Declaration of response codes
    Private Const CODE_DIGITAL_OUT_SINGLE As Integer = 100
    Private Const CODE_DIGITAL_OUT_ALL As Integer = 101
    Private Const CODE_DIGITAL_IN_SINGLE As Integer = 200
    Private Const CODE_DIGITAL_IN_ALL As Integer = 201
    Private Const CODE_ANALOG_OUT_SINGLE As Integer = 300
    Private Const CODE_ANALOG_OUT_ALL As Integer = 301
    Private Const CODE_ANALOG_IN_SINGLE As Integer = 400
    Private Const CODE_ANALOG_IN_ALL As Integer = 401



    Public Function Cmd_setAllDigital() As ServerResponse
        mK8055.setAllDigitalOutputs()

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_clearAllDigital() As ServerResponse
        mK8055.clearAllDigtalOutputs()

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_setDigital(channel As DigitalOutput) As ServerResponse
        mK8055.setDigitalOutput(channel.toInt())

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_clearDigital(channel As DigitalOutput) As ServerResponse
        mK8055.clearDigitalOutput(channel.toInt())

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_toggleDigital(channel As DigitalOutput) As ServerResponse
        mK8055.toggleDigitalOutput(channel.toInt())

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_toggleAllDigital() As ServerResponse
        mK8055.toggleAllDigitalOutputs()

        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_invertAllDigital() As ServerResponse
        Return Cmd_toggleAllDigital()
    End Function




    Public Function Cmd_invertDigital(channel As DigitalOutput) As ServerResponse
        Return Cmd_toggleDigital(channel)
    End Function




    Public Function Cmd_getAllDigital() As ServerResponse
        Return New ServerResponse(CODE_DIGITAL_OUT_ALL,
                                  mK8055.getDigitalOutputLine())
    End Function




    Public Function Cmd_getInput(channel As DigitalInput) As ServerResponse
        Return New ServerResponse(CODE_DIGITAL_IN_SINGLE,
                                    mK8055.getDigitalInput(channel.toInt).ToString())
    End Function




    Public Function Cmd_GetAllInputs() As ServerResponse
        Return New ServerResponse(CODE_DIGITAL_IN_ALL,
                                    mK8055.getDigitalInputs())
    End Function




    Public Function Cmd_setAnalogOutput(channel As AnalogOutput) As ServerResponse
        mK8055.setAnalogOutputToMax(channel.toInt())

        Return New ServerResponse(CODE_ANALOG_OUT_SINGLE,
                                   mK8055.getAnalogOutput(channel.toInt).ToString())
    End Function




    Public Function Cmd_setAnalogOutputTo(channel As AnalogOutput, value As AnalogValue) As ServerResponse
        mK8055.setAnalogOutputToValue(channel.toInt(), value.toInt())

        Return New ServerResponse(CODE_ANALOG_OUT_SINGLE,
                                   mK8055.getAnalogOutput(channel.toInt()).ToString())
    End Function




    Public Function Cmd_clearAnalogOutput(channel As AnalogOutput) As ServerResponse
        mK8055.setAnalogOutputToValue(channel.toInt(), 0)

        Return New ServerResponse(CODE_ANALOG_OUT_SINGLE,
                                   mK8055.getAnalogOutput(channel.toInt).ToString())
    End Function




    Public Function Cmd_getAnalogOutput(channel As AnalogOutput) As ServerResponse
        Return New ServerResponse(CODE_ANALOG_OUT_SINGLE,
                                   mK8055.getAnalogOutput(channel.toInt).ToString())
    End Function




    Public Function Cmd_getAllAnalogOutputs() As ServerResponse
        Return New ServerResponse(CODE_ANALOG_OUT_ALL,
                                   mK8055.getAnalogOutputs())
    End Function




    Public Function Cmd_getAnalogInput(channel As AnalogInput) As ServerResponse
        Return New ServerResponse(CODE_ANALOG_IN_ALL,
                                   mK8055.ReadAnalogInput(channel.toInt).ToString())
    End Function




    Public Function Cmd_getAllAnalogInputs() As ServerResponse
        Return New ServerResponse(CODE_ANALOG_IN_ALL,
                                   mK8055.getAllAnalogInputs())
    End Function
#End Region
End Class