﻿Option Explicit On
Option Strict On

Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports System.Net
Imports Microsoft.VisualBasic.FileIO


''' <summary>
''' ClientConnection() represents a single connection between the server and the client.
''' The class uses an internal background thread that runs the actual connection.
''' </summary>
''' <remarks>Copyright (c) 2015   Vanderhaegen Bart</remarks>
Public Class ClientConnection

    ' Properties
    Public Property clientSocket As TcpClient
    Public Property sessionID As Integer
    Public Property username As String
    Public Property password As String
    Public Property supportsTelnet As Boolean

    ' Delegate properties
    Public Property chSignalFeedback As Action(Of Integer, ClientConnection.Signals)
    Public Property chExecuteCommand As Func(Of Integer, String, String, String)
    Public Property chFeedbackLogger As Action(Of String, String, String)
    Public Property chFeedbackLog As Action(Of String, String, String)

    ' Connection state variables
    Private isAuthenticated As Boolean
    Private connectionEnabled As Boolean
    Private userNameOk As Boolean
    Private passwordOk As Boolean
    Private cancelTask As Boolean
    Private clientIP As String = String.Empty

    ' Internal variables
    Private networkStream As NetworkStream  ' the network connection
    Private ctThread As Thread              ' the background thread



    ' Enumeration with possible feedback signals
    Public Enum Signals
        USER_LOGGED_IN
        CONNECTION_TERMINATED
    End Enum





    ''' <summary>
    ''' isThreadRunning() checks if the thread with the connection is still running.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function isThreadRunning() As Boolean
        If ctThread Is Nothing Then
            Return False
        Else
            Return ctThread.IsAlive()
        End If
    End Function




    ''' <summary>
    ''' isConnectionEnabled() checks if the client is still active.
    ''' </summary>
    ''' <returns>A boolean that indicates if the serverconnection is still active and
    ''' listening for new commands.</returns>
    ''' <remarks></remarks>
    Public Function isConnectionEnabled() As Boolean
        Return connectionEnabled
    End Function




    ''' <summary>
    ''' Class constructor.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
    End Sub




    ''' <summary>
    ''' start() will execute the client connection. This should not be done until
    ''' all delegate properties are configured.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub start()
        If isThreadRunning() Then
            ' Thread is already running: throw a new exception
            Throw New Exception("The client thread is already running")
        Else
            ' Execute the thread
            clientIP = getIP()
            Me.connectionEnabled = True
            ctThread = New Threading.Thread(AddressOf connectionEstablished)
            ctThread.IsBackground = True
            cancelTask = False
            ctThread.Start()
        End If
    End Sub



    ''' <summary>
    ''' terminate() will stop the client connection.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub terminate()
        connectionEnabled = False

        If clientSocket IsNot Nothing Then
            clientSocket.Close()
            clientSocket = Nothing
        End If
    End Sub




    ''' <summary>
    ''' getIP() gets the IP address for the connected client
    ''' </summary>
    ''' <returns>The IP address in a string format.</returns>
    ''' <remarks></remarks>
    Public Function getIP() As String
        If clientSocket IsNot Nothing Then
            Return DirectCast(clientSocket.Client.RemoteEndPoint, IPEndPoint).Address.ToString()
        Else
            Return String.Empty
        End If
    End Function




    ''' <summary>
    ''' sendLog() will send a feedback message to the assigned delegate function. The message
    ''' will appear in the list of the main screen
    ''' </summary>
    ''' <param name="message">The message to be transmitted</param>
    ''' <remarks></remarks>
    Private Sub sendLog(message As String)
        chFeedbackLogger.Invoke(sessionID.ToString(), clientIP, message)
    End Sub




    ''' <summary>
    ''' bufferToString() converts the buffer array with the incoming data to a String. The function
    ''' automatically strips all Telnet control commands.
    ''' </summary>
    ''' <param name="bufferArray">The array containing the buffer data</param>
    ''' <param name="numberOfBytes">The number of received bytes</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function bufferToString(bufferArray() As Byte, numberOfBytes As Integer) As String
        ' Only executes extended processing if the server is in Telnet-mode
        If supportsTelnet Then
            ' Convert our data to a list for easier access
            Dim bufferList As List(Of Byte) = bufferArray.ToList()

            ' Remote the part of the list that is empty &H00.
            bufferList.RemoveRange(numberOfBytes, bufferList.Count - numberOfBytes)

            ' Strip all Telnet control commands
            Dim i As Integer
            While i < bufferList.Count - 1
                If bufferList(i) = &HFF Then
                    bufferList.RemoveRange(i, 3)
                Else
                    i += 1
                End If
            End While

            ' Converts the buffer back to an array
            bufferArray = bufferList.ToArray()
        End If

        ' Convert the result to an ASCII string
        Return System.Text.Encoding.ASCII.GetString(bufferArray, 0, bufferArray.Count)
    End Function




    ''' <summary>
    ''' connectionEstablished() runs the background thread.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub connectionEstablished()

        Dim myReadBuffer(1024) As Byte
        Dim dataFromClient As String

        Dim response As StringBuilder = New StringBuilder
        Dim myResponse As String = String.Empty
        Dim numberOfBytesToRead As Integer = 0

        Try
            networkStream = clientSocket.GetStream()

            ' Show initial server response
            sendResponse("+OK Welcome to K8055 server")
            sendLog("Client connected")

            ' Loop as long as the connection is active
            While (connectionEnabled)
                Try
                    If networkStream.CanRead() Then
                        ' Get the number of bytes that are available to be read.
                        numberOfBytesToRead = networkStream.Read(myReadBuffer, 0, myReadBuffer.Length)

                        If numberOfBytesToRead = 0 Then
                            ' Zero read bytes means the connection has been broken.
                            connectionEnabled = False

                        Else
                            ' One or more bytes have been read. Get the read data.
                            dataFromClient = bufferToString(myReadBuffer, numberOfBytesToRead)

                            ' Send back backspace character
                            If supportsTelnet AndAlso dataFromClient = vbBack Then
                                sendResponseLine(" " & vbBack)
                                myResponse = myResponse & dataFromClient
                            Else
                                myResponse = myResponse & dataFromClient
                            End If

                            ' Wait until a line break (carriage return + line feed) has been received
                            If myResponse.Contains(vbCrLf) Then
                                ' Split the received string into an array of strings
                                Dim lines As String() = myResponse.Split(
                                    New String() {vbCrLf}, StringSplitOptions.None)

                                ' Process the data if at least one line has been received
                                If lines.Count > 0 Then
                                    processData(lines)
                                    myResponse = String.Empty
                                End If 'lines.Count > 0 Then
                            End If 'myResponse.Contains(vbCrLf) Then
                        End If 'numberOfBytesToRead = 0 Then
                    End If 'networkStream.CanRead() Then

                Catch ex As Exception
                    connectionEnabled = False
                End Try

            End While
        Catch ex As Exception
            Debug.Print("Server error: " & ex.Message())
        End Try

        terminate()
        sendLog("Client disconnected")
        chSignalFeedback.Invoke(sessionID, Signals.CONNECTION_TERMINATED)
    End Sub




    ''' <summary>
    ''' sendResponse() will send a response back to the client. A vbcrlf character
    ''' will be appended automatically (carriage return + newline).
    ''' </summary>
    ''' <param name="response">The response to be transmitted.</param>
    ''' <remarks></remarks>
    Private Sub sendResponse(ByVal response As String)
        sendResponseLine(response & vbCrLf)
    End Sub




    ''' <summary>
    ''' sendResponse() will send a response back to the client. A vbcrlf character
    ''' will be appended automatically (carriage return + newline).
    ''' </summary>
    ''' <param name="response">The response to be transmitted.</param>
    ''' <remarks></remarks>
    Private Sub sendResponseLine(ByVal response As String)
        Dim sendBytes As [Byte]()
        sendBytes = Encoding.ASCII.GetBytes(response)
        networkStream.Write(sendBytes, 0, sendBytes.Length)
        networkStream.Flush()
    End Sub




    ''' <summary>
    ''' splitLine() will convert the received line from the client into an array of String objects.
    ''' It uses a space-character as the decimal separator.
    ''' </summary>
    ''' <param name="line"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function splitLine(line As String) As String()
        Dim fields() As String

        Using rdr As New IO.StringReader(line)
            Using parser As New TextFieldParser(rdr)
                parser.TextFieldType = FieldType.Delimited
                parser.Delimiters = New String() {" "}
                parser.HasFieldsEnclosedInQuotes = True
                fields = parser.ReadFields()
            End Using
        End Using

        If fields IsNot Nothing AndAlso fields.Length > 0 Then
            ' Make sure the first field is always in uppercase
            fields(0) = fields(0).ToUpper()
        Else
            ' Make sure there is always a non-null response
            fields = {String.Empty}
        End If

        Return fields
    End Function




    ''' <summary>
    ''' processBackspaceCharacters() will process the string for backspace characters (for
    ''' MS Telnet Client compatibility).
    ''' </summary>
    ''' <param name="command">the command string</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function processBackspaceCharacters(command As String) As String
        Dim response As New StringBuilder()

        ' Loop through each char:
        ' Append when non-backspace, remove last char when backspace
        For Each c As Char In command.ToCharArray()
            If c = vbBack Then
                If response.Length() > 0 Then
                    response.Remove(response.Length - 1, 1)
                End If
            Else
                response.Append(c)
            End If
        Next

        Return response.ToString()
    End Function




    ''' <summary>
    ''' processReceivedLine() will process the received line.
    ''' </summary>
    ''' <param name="command">The command string to process</param>
    ''' <remarks></remarks>
    Private Sub processReceivedLine(ByVal command As String)
        Dim doContinue As Boolean

        ' Ensure MS Telenet compatibility: process backspace characters
        command = processBackspaceCharacters(command)

        ' Decode command and parameters
        Dim myFields() As String = splitLine(command)
        Dim currentInstruction As String = myFields(0)
        Dim currentParameter As String = getCompleteParameterLine(command)
        sendLog(String.Format("Command: {0}", currentInstruction))

        ' Check if the received commands correspond with any of the hardcoded commands 
        ' (username, password, quit).
        Select Case currentInstruction
            Case "USER"
                ' USER <username>
                userNameOk = (currentParameter = username)
                passwordOk = False
                sendResponse("+OK 001 Username received")


            Case "PASS"
                ' PASS <password>
                If userNameOk Then
                    If currentParameter = password Then
                        passwordOk = True
                        isAuthenticated = True
                        sendResponse("+OK 002 User authenticated")
                        sendLog("Client authenticated")
                        chSignalFeedback.Invoke(sessionID, Signals.USER_LOGGED_IN)
                    Else
                        passwordOk = False
                        sendResponse("-ERR Authentication failed")
                        sendLog("Client authentication failed (wrong password)")
                    End If
                Else
                    passwordOk = False
                    sendResponse("-ERR Authentication failed")
                    sendLog("Client authentication failed (wrong username)")
                    connectionEnabled = False
                End If


            Case "QUIT"
                ' QUIT
                connectionEnabled = False

            Case String.Empty
                sendResponse("-ERR Unknown command")

            Case Else
                ' An unknown command was entered. Set doContinue to true so
                ' the function will be executed.
                doContinue = True
        End Select



        If doContinue Then
            If isAuthenticated Then
                ' Only continues if the user is already authenticated
                Dim res As String
                ' Calls the delegate that is in charge with the execution of commands
                res = chExecuteCommand.Invoke(sessionID,
                                         currentInstruction,
                                         currentParameter)
                ' Send the response back to the client
                sendResponse(res)
            Else
                ' The user has not been authenticated yet
                sendResponse("-ERR Please authenticate first")
            End If
        End If
    End Sub




    ''' <summary>
    ''' getParameter() gets the entered parameter after a command.
    ''' </summary>
    ''' <param name="command">The complete command line</param>
    ''' <returns>All of the parameters</returns>
    ''' <remarks></remarks>
    Private Function getCompleteParameterLine(ByVal command As String) As String
        command = command.Trim()

        Dim locOfFirtSpace As Integer = InStr(command, " ")

        If locOfFirtSpace = 0 Then
            Return String.Empty
        Else
            Return Mid(command, locOfFirtSpace + 1, command.Length - locOfFirtSpace)
        End If
    End Function




    ''' <summary>
    ''' processData() will execute the received lines one by one.
    ''' </summary>
    ''' <param name="dataLines">An array with all of the datalines.</param>
    ''' <remarks></remarks>
    Private Sub processData(ByVal dataLines() As String)
        If dataLines IsNot Nothing Then
            For i As Integer = 0 To UBound(dataLines) - 1
                processReceivedLine(dataLines(i))
            Next
        End If

        For Each line As String In dataLines
            'If line.Trim() <> String.Empty Then
            ' processReceivedLine(line)
            'End If
        Next
    End Sub

End Class