﻿Option Explicit On
Option Strict On

Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports System.Net

''' <summary>
''' The Server Class represents the server listener. There should only be one instance
''' of this class.
''' </summary>
''' <remarks>Copyright (c) 2015   Vanderhaegen Bart</remarks>
Public Class Server

    ' Server Definitions
    Private Const MAX_CLIENTS As Integer = 10

    ' Shared Class variables
    Private Shared myServer As Server

    ' Class variables
    Private clientList As ClientConnection
    Private serverThread As Threading.Thread
    Private serverSocket As TcpListener
    Private serverRunning As Boolean
    Private activeClients As New List(Of ClientConnection)
    Private clientsLock As New Object()

    ' Class properties
    Public Property server_user As String
    Public Property server_pass As String
    Public Property server_port As Integer
    Public Property supportsTelnet As Boolean
    Public Property interpreter As ServerFunctions
    Public Property dlgCommand As Func(Of Integer, String, String, String)  'Server.dlgExecuter

    ' Definition of events
    Public Event server_started()
    Public Event server_stopped(reason As String)
    Public Event client_connected(clientID As Integer, IP As String)
    Public Event client_disconnected(clientID As Integer)
    Public Event instruction_received(clientID As Integer, command As String, parameter As String)
    Public Event client_signal_received(clientID As Integer, eventType As ClientConnection.Signals)
    Public Event log_received(clientID As String, IP As String, message As String)



    ' Singleton implementation of the class
    Public Shared Function getServer() As Server
        If myServer IsNot Nothing Then
            Return myServer
        Else
            Return New Server
        End If
    End Function




    ''' <summary>
    ''' isRunning() checks if the server is already running
    ''' </summary>
    ''' <returns>A boolean to indicate the server state</returns>
    ''' <remarks></remarks>
    Public Function isRunning() As Boolean
        Return (serverThread IsNot Nothing AndAlso serverThread.IsAlive())
    End Function




    ''' <summary>
    ''' start() will execute the background server
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub start()
        serverRunning = True
        serverThread = New Threading.Thread(AddressOf background_Server)
        serverThread.IsBackground = True
        serverThread.Start()
    End Sub




    ''' <summary>
    ''' stopClientConnections() will close all opened connections
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub stopClientConnections()
        Try
            For Each connection As ClientConnection In activeClients
                connection.terminate()
            Next
        Catch ex As Exception
            ' Under certain circumstances the function triggers an error. We don't
            ' need any feedback for this right now.
            ' TODO: improve
        End Try
    End Sub




    ''' <summary>
    ''' killInactiveInstances() loops through all open active client objects and
    ''' kills the ones that are no longer active.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub killInactiveInstances()
        SyncLock clientsLock
            For i As Integer = activeClients.Count - 1 To 0 Step -1
                ' Delete reference to unused instances
                If Not activeClients(i).isConnectionEnabled Then
                    activeClients.RemoveAt(i)
                End If
            Next
        End SyncLock
    End Sub




    ''' <summary>
    ''' stop() will terminate the server
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub [stop]()
        serverRunning = False
        quitServer()
        stopClientConnections()
    End Sub




    ''' <summary>
    ''' quitServer() will close the server socket
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub quitServer()
        If serverSocket IsNot Nothing Then
            serverSocket.Stop()
        End If
    End Sub




    ''' <summary>
    ''' log_received() sends a system logging message (without sessionID or IP) 
    ''' to the log list.
    ''' </summary>
    ''' <param name="text">The message to send</param>
    ''' <remarks></remarks>
    Private Sub log_system(text As String)
        RaiseEvent log_received(String.Empty,
                                String.Empty,
                                text)
    End Sub




    ''' <summary>
    ''' log_client() sends a complete logging message to the log list.
    ''' </summary>
    ''' <param name="clientID">The id of the client that triggered the event</param>
    ''' <param name="IP">The IP address of the client</param>
    ''' <param name="message">The message to send</param>
    ''' <remarks></remarks>
    Private Sub log_client(clientID As String, IP As String,
                           message As String)

        RaiseEvent log_received(clientID,
                                 IP,
                                 message)
    End Sub




    ''' <summary>
    ''' background_Server() runs the background server process
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub background_Server()
        Dim stopReason As String = "Server shutdown by user"
        log_system("Server Started")

        Try
            ' Create a listening socket on all local IP addresses
            serverSocket = New TcpListener(IPAddress.Any, server_port)
            Dim sessionID As Integer
            Dim serverRunning As Boolean = False


            Try
                serverSocket.Start()    ' Enable server
                RaiseEvent server_started()
                serverRunning = True
            Catch ex As SocketException
                If ex.SocketErrorCode = SocketError.AddressAlreadyInUse Then
                    stopReason = String.Format("Unable to execute server: TCP port {0} is already in use", server_port)
                Else
                    stopReason = String.Format("Unable to execute server: {0} {1}", ex.Message, ex.ErrorCode)
                End If

            End Try

            While (serverRunning)
                ' Calculate new sessionID
                sessionID += 1

                ' Wait for new connection and accept it
                Dim clientSocket As TcpClient =
                     serverSocket.AcceptTcpClient()

                If countActiveSessions() < MAX_CLIENTS Then
                    ' Create new client information
                    Dim client As New ClientConnection With {
                        .sessionID = sessionID,
                        .clientSocket = clientSocket,
                        .username = server_user,
                        .password = server_pass,
                        .supportsTelnet = supportsTelnet,
                        .chSignalFeedback = AddressOf callback_clientSignal,
                        .chFeedbackLog = AddressOf log_client,
                        .chFeedbackLogger = AddressOf log_client,
                        .chExecuteCommand = dlgCommand
                    }

                    ' Execute the client background connection
                    client.start()

                    ' Add client to our list with all active clients
                    SyncLock clientsLock
                        activeClients.Add(client)
                    End SyncLock


                    ' Show new connection in list
                    RaiseEvent client_connected(sessionID, client.getIP())

                Else
                    sendResponse(clientSocket, "-ERR All server connections are in use")

                    clientSocket.Close()
                    clientSocket = Nothing
                End If

            End While


        Catch ex As Exception
            If serverRunning Then
                stopReason = String.Format("Unexpected server shutdown: {0}", ex.ToString)
            Else
                stopReason = "Stopped on user request"
            End If
        End Try

        ' server disconnected
        disconnectClients()
        log_system(String.Format("Server Stopped ({0})", stopReason))
        RaiseEvent server_stopped(stopReason)
    End Sub




    ''' <summary>
    ''' sendResponse() will send an initial response back to the client (in case of server errors)
    ''' </summary>
    ''' <param name="sMessage"></param>
    ''' <remarks></remarks>
    Private Sub sendResponse(oSocket As TcpClient, sMessage As String)
        Try
            Dim networkStream As NetworkStream = oSocket.GetStream()

            ' Create an array of bytes and send over the network
            Dim sendBytes As [Byte]()
            sendBytes = Encoding.ASCII.GetBytes("-ERR all server connections are in use" & vbCrLf)
            networkStream.Write(sendBytes, 0, sendBytes.Length)
            networkStream.Flush()
        Catch ex As Exception
            ' Transmitting error occured. We don't care (connection was to be closed anyway)
        End Try
    End Sub




    ''' <summary>
    ''' countActiveSessions() will count all active connections that are still connected
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function countActiveSessions() As Integer
        Dim res As Integer = 0
        For Each client As ClientConnection In activeClients
            If client IsNot Nothing AndAlso client.isThreadRunning Then
                res += 1
            End If
        Next
        Return res
    End Function




    ''' <summary>
    ''' disconnectClients() will disconnect all active clients from the server
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub disconnectClients()
        ' Close all open client connections
        For Each connection As ClientConnection In activeClients
            connection.terminate()
            connection = Nothing
        Next

        ' Delete all open connections
        activeClients.Clear()
    End Sub




    ''' <summary>
    ''' callback_clientSignal() will trigger a client message on the GUI
    ''' </summary>
    ''' <param name="clientID">The client that triggered the message</param>
    ''' <param name="message">The type of message</param>
    ''' <remarks></remarks>
    Private Sub callback_clientSignal(clientID As Integer,
                                      message As ClientConnection.Signals)

        If message = ClientConnection.Signals.CONNECTION_TERMINATED Then
            killInactiveInstances()
            RaiseEvent client_disconnected(clientID)
        End If


        RaiseEvent client_signal_received(clientID, message)
    End Sub

End Class